#!/usr/bin/python3

# Copyright (C) 2021 Canonical Ltd.
# Author: Haw Loeung <haw.loeung@canonical.com>

import base64
import logging
import os
import requests
import socket
import time
import uuid

from urllib.parse import urljoin

CHECK_HEADERS = {
    'Host': 'archive.ubuntu.com',
    'User-agent': 'ubuntu-repository-cache/healthcheck',
    'Accept-Encoding': '*',
}
CHECK_HOST = 'http://localhost/'
CHECK_TIMEOUT = 5
CHECK_URLS = ['/ubuntu/dists/focal/InRelease', '/ubuntu/pool/main/']

# By default, check the flag file in the current path as the
# script. Otherwise, we'll want to make this a template and then have
# the charm use unitdata.kv().get('apache-root') or similar.
DISABLE_FLAG = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'health-check-disabled.flag')

LOG_FILE = '/var/log/ubuntu-repository-cache/health-check.log'

METRIC_NAME = 'ubuntu_repository_cache_health'


def generate_uuid(uid=uuid.uuid4().bytes):
    uid = base64.b64encode(uid, b'pa')[0:8].decode('utf-8')
    return uid


def render_influx(metric_name, labels, value):
    line_fmt = "{metric_name} value={value}"
    if labels:
        line_fmt = "{metric_name},{labels} value={value}"
    return line_fmt.format(metric_name=metric_name, labels=labels, value=value)


def send_to_influx(metrics, send_to='127.0.0.1:8094'):
    (host, port) = send_to.split(':')
    port = int(port)
    try:
        s = socket.socket()
        s.settimeout(2)
        s.connect((host, port))
        with s.makefile(mode='w') as dest:
            dest.write(metrics)
    except Exception as err:
        logging.warning('Failed to send metrics: %s', err)
        # We don't care if it fails to send metrics.
        pass


def check_url(url):
    start_time = time.time()

    count = 0
    while count < 3:
        count += 1

        status_code = 0
        start_time_url = time.time()
        msg = ''
        try:
            check_url = urljoin(CHECK_HOST, url)
            resp = requests.head(check_url, headers=CHECK_HEADERS, timeout=CHECK_TIMEOUT)
            status_code = resp.status_code
            if resp.ok:
                duration = time.time() - start_time_url
                logging.info('OK: #%s: %s (%.3f)', count, url, duration)
                break
            else:
                resp.raise_for_status()
        except requests.exceptions.HTTPError as err:
            duration = time.time() - start_time_url
            logging.error('ERROR: #%s: %s: %s (%.3f)', count, url, err, duration)
            msg = err
        except (
            requests.exceptions.ConnectTimeout,
            requests.exceptions.ConnectionError,
            requests.exceptions.Timeout,
        ) as err:
            duration = time.time() - start_time_url
            logging.error('ERROR: #%s: %s: %s (%.3f)', count, url, err, duration)
            msg = err
        except Exception as err:
            duration = time.time() - start_time_url
            # We don't want to show exactly what the error is, but we
            # do want to log it.
            logging.critical('ERROR: #%s: %s: %s (%.3f)', count, url, err, duration)
            msg = 'Unknown error'

    return status_code, time.time() - start_time, msg


def main(disabled_flag=DISABLE_FLAG):
    # logging lines include a unique identifier so it's easier to grep
    # and group all related lines. 'pa' are alternate characters
    # replacing '+' with 'p' and '/' with 'a'.
    uid = generate_uuid()
    remote_addr = os.environ.get('REMOTE_ADDR', '-')
    logging.basicConfig(
        filename=LOG_FILE,
        format='%(asctime)s {} [{}]: %(message)s'.format(remote_addr, uid),
        datefmt='%Y-%m-%d %H:%M:%S',
        level=logging.INFO,
    )

    if os.path.exists(disabled_flag):
        print('Status: 503 Service Unavailable')
        print('Content-type: text/html\n')
        disabled_time = os.stat(disabled_flag).st_mtime
        print('Currently disabled for maintenance since {}'.format(time.ctime(int(disabled_time))), flush=True)
        logging.info('DISABLED since %.3f', disabled_time)
        # Rather than 0 here, we want to make maintenance/disabled
        # units stand out a little in the grafana graphs, so 5.03s with
        # the average being ~200-300 ms to the backends.
        send_to_influx(render_influx(METRIC_NAME, 'status=DISABLED', 5.03))
        return

    # Ensure ubuntu_active exists. Useful for initial provisioning
    # where it's still being sync'ed.
    apache_root = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    if not os.path.exists(os.path.join(apache_root, 'data', 'ubuntu_active')):
        print('Status: 503 Service Unavailable')
        print('Content-type: text/html\n')
        print('Currently provisioning or missing metadata', flush=True)
        logging.warning('DISABLED ubuntu_active missing')
        # Rather than 0 here, we want to make disabled units stand out
        # a little in the grafana graphs, so 4.04s with the average
        # being ~200-300 ms to the backends.
        send_to_influx(render_influx(METRIC_NAME, 'status=DISABLED', 4.04))
        return

    start_time = time.time()

    successes = {}
    failures = {}
    for url in CHECK_URLS:
        res = check_url(url)
        if res[0] == 200:
            successes[url] = res
        else:
            failures[url] = res

    metrics = []
    if failures:
        status = '503 Service Unavailable'
        print('Status: {}'.format(status))
        print('Content-type: text/html\n')
        for url, val in failures.items():
            status_code = val[0]
            duration = val[1]
            err = val[2]
            print('ERROR: {}: {} ({:.3f}s)'.format(url, err, duration))
            labels = 'url={},status=FAIL,status_code={}'.format(url, status_code)
            metrics.append(render_influx(METRIC_NAME, labels, duration))
    else:
        status = '200 OK'
        print('Status: {}'.format(status))
        print('Content-type: text/html\n')

    # Either way, output all successful checks.
    for url, val in successes.items():
        status_code = val[0]
        duration = val[1]
        print('OK: {} ({:.3f}s)'.format(url, duration))
        labels = 'url={},status=SUCCESS,status_code={}'.format(url, status_code)
        metrics.append(render_influx(METRIC_NAME, labels, duration))

    # Flush the buffer to ensure we're not holding up
    # request/connection on sending metrics to influx.
    print('', flush=True)

    duration = time.time() - start_time
    logging.info('%s (%.3f)', status, duration)

    for metric in metrics:
        send_to_influx(metric)

    return


if __name__ == '__main__':
    main()
