'''Functions related to apache use within this charm'''

import multiprocessing
import os
import subprocess
import json

from charmhelpers import fetch

from charmhelpers.core import (
    hookenv,
    host,
    templating,
    unitdata,
)
from . import util


HEALTH_CHECK_LOG = '/var/log/ubuntu-repository-cache/health-check.log'
# The user apache2 runs as needs access to write to the logs
HEALTH_CHECK_LOG_OWNER = 'www-data:adm'
LOG = hookenv.log
SERVICE = 'apache2'


@util.run_as_user('root')
def install():
    '''Perform the install/config of apache and disable the default site'''

    LOG('Installing apache', hookenv.INFO)
    fetch.apt_install(SERVICE, fatal=True)
    stop()
    subprocess.check_call(['a2enmod', 'expires'])
    subprocess.check_call(['a2enmod', 'proxy'])
    subprocess.check_call(['a2enmod', 'proxy_http'])
    subprocess.check_call(['a2enmod', 'headers'])
    subprocess.check_call(['a2dismod', 'mpm_event'])
    subprocess.check_call(['a2enmod', 'mpm_worker'])
    subprocess.check_call(['a2dissite', '000-default'])

    config = hookenv.config()
    if config['enable_healthcheck']:
        subprocess.check_call(['a2enmod', 'cgid'])
    if config['remoteip_logging']:
        subprocess.check_call(['a2enmod', 'remoteip'])

    start()


@util.run_as_user('root')
def start(start_only=False):
    '''Start the service and remove the override file blocking startup'''

    override_path = '/etc/init/apache2.override'
    if os.path.exists(override_path):
        os.remove('/etc/init/apache2.override')

    if not service_running():
        LOG('Starting apache', hookenv.INFO)
        host.service_start(SERVICE)
    elif not start_only:
        LOG('Reloading apache', hookenv.INFO)
        host.service_reload(SERVICE)
    else:
        LOG('apache already running', hookenv.INFO)


@util.run_as_user('root')
def stop():
    '''Stop the service and block startup at boot'''

    LOG('Stopping apache', hookenv.INFO)
    with open('/etc/init/apache2.override', 'w+') as override_file:
        override_file.write('manual\n')

    host.service_stop(SERVICE)


@util.run_as_user('root')
def service_running():
    '''Return True is the service is running'''

    return host.service_running(SERVICE)


def tune_mpm_configs(config):
    cpu_count = multiprocessing.cpu_count()

    if not config['startservers']:
        config['startservers'] = 1

    if not config['serverlimit']:
        # Benchmarking shows single process multiple threads performs better
        # than multiple processes multiple threads.
        # Unfortunately, we need more than one server process to allow for
        # graceful restarts. See LP#1918211.
        config['serverlimit'] = 2

    if not config['threadsperchild']:
        # We could take into consideration how much memory the system
        # has but the ratio of vCPU to RAM between public clouds seems
        # good enough so this is not needed.
        config['threadsperchild'] = 256 * cpu_count
        # Cap at 2560. It's been observed in AWS, even though there
        # are 36 vCPUs, VMs perform badly with over 2500 workers
        # (likely squid-deb-proxy bottleneck). But let's cap it here,
        # someone can override it by setting the
        # apache2_mpm_threadlimit charm config.
        if (config['threadsperchild'] * config['serverlimit']) > 2560:
            config['threadsperchild'] = int(2560 / config['serverlimit'])

    if not config['threadlimit']:
        # Hard limit should match per child limit.
        config['threadlimit'] = config['threadsperchild']

    if not config['maxrequestworkers']:
        # Calculate max. workers based on serverlimit and threadsperchild.
        config['maxrequestworkers'] = config['serverlimit'] * config['threadsperchild']

    if not config['minsparethreads']:
        # Use 50% of max. workers.
        config['minsparethreads'] = int(config['maxrequestworkers'] * 0.5)

    if not config['maxsparethreads']:
        config['maxsparethreads'] = config['maxrequestworkers']

    return config


def create_mpm_workerfile(config=None):
    '''Create the multi-processing module (MPM) configuration'''

    reload_required = False

    if not config:
        config = hookenv.config()

    mpm_configs = [
        'maxconnectionsperchild',
        'maxrequestworkers',
        'maxsparethreads',
        'minsparethreads',
        'serverlimit',
        'startservers',
        'threadlimit',
        'threadsperchild',
    ]
    mpm_context = {}
    for k in mpm_configs:
        cfkey = 'apache2_mpm_{}'.format(k)
        mpm_context[k] = config[cfkey]
        if config.changed(cfkey):
            reload_required = True

    # We'll only proceed when any one of the MPM configs have changed.
    if not reload_required:
        return reload_required

    reload_required = True

    mpm_context = tune_mpm_configs(mpm_context)
    conf_mpm_worker = '/etc/apache2/conf-available/000mpm-worker.conf'
    templating.render('apache2/mpm_worker.template', conf_mpm_worker, mpm_context)
    subprocess.check_call(['a2enconf', '000mpm-worker'])
    mods_mpm_worker = '/etc/apache2/mods-available/mpm_worker.conf'
    if not os.path.exists(mods_mpm_worker):
        os.symlink(conf_mpm_worker, mods_mpm_worker)
    elif not os.path.islink(mods_mpm_worker):
        os.unlink(mods_mpm_worker)
        os.symlink(conf_mpm_worker, mods_mpm_worker)

    return reload_required


def create_security(config=None):
    '''Create the security configuration'''

    reload_required = False

    if not config:
        config = hookenv.config()

    security_configs = ['server_signature', 'server_tokens', 'trace_enabled']
    security_context = {}
    for k in security_configs:
        cfkey = 'apache2_{}'.format(k)
        security_context[k] = config[cfkey]
        if config.changed(cfkey):
            reload_required = True

    # We'll only proceed when any one of the MPM configs have changed.
    if not reload_required:
        return reload_required

    reload_required = True

    templating.render('apache2/security.template', '/etc/apache2/conf-available/security.conf', security_context)
    subprocess.check_call(['a2enconf', 'security'])
    return reload_required


def create_metadata_site(config=None, apache2_conf_path='/etc/apache2'):
    '''Create the site config for serving (or proxying) repository metadata'''

    reload_required = False

    if not config:
        config = hookenv.config()

    apache_context = {}
    apache_context['DocumentRoot'] = unitdata.kv().get('apache-root')
    apache_context['Sync_Host'] = config['sync-host']
    apache_context['Display_Host'] = config['display-host']
    apache_context['Path_Base'] = config['path-base']
    metaproxy_conf = '# No failover configured'
    failover_host = util.get_failover_host()
    if util.in_failover() and failover_host:
        uri = 'http://%s:80' % failover_host
        metaproxy_conf = (
            'ProxyRemote      http://{disp}/ {uri}\n'
            '\tProxyPass        / http://{disp}/\n'
            '\tProxyPassReverse / http://{disp}/'
        ).format(uri=uri, disp=config['display-host'])
    apache_context['MetadataProxy'] = metaproxy_conf
    apache_context['Enable_Healthcheck'] = config['enable_healthcheck']

    apache_config_file = os.path.join(apache2_conf_path, 'sites-available', 'archive_ubuntu_com.conf')
    cur_conf = ''
    if os.path.exists(apache_config_file):
        with open(apache_config_file) as f:
            cur_conf = f.read()
    new_conf = templating.render('apache2/archive_ubuntu_com.conf', apache_config_file, apache_context)

    if new_conf != cur_conf:
        reload_required = True
        if not os.path.exists(os.path.join(apache2_conf_path, 'sites-enabled', 'archive_ubuntu_com.conf')):
            subprocess.call(['a2ensite', 'archive_ubuntu_com'])

    return reload_required


@util.run_as_user('root')
def configure_remoteip_logging(config=None, apache2_conf_path='/etc/apache2'):
    reload_required = False

    if not config:
        config = hookenv.config()

    if config['remoteip_logging']:
        apache_config_file = os.path.join(apache2_conf_path, 'conf-available', 'remoteip.conf')
        cur_conf = ''
        if os.path.exists(apache_config_file):
            with open(apache_config_file) as f:
                cur_conf = f.read()
        new_conf = templating.render('apache2/remoteip.conf', apache_config_file, {})

        if new_conf != cur_conf:
            reload_required = True
            if not os.path.exists(os.path.join(apache2_conf_path, 'conf-enabled', 'remoteip.conf')):
                subprocess.check_call(['a2enconf', 'remoteip'])

    if config.changed('remoteip_logging'):
        LOG('Apache remoteip_logging changed')
        if config['remoteip_logging']:
            subprocess.check_call(['a2enmod', 'remoteip'])
            LOG('Enabled remoteip logging')
        else:
            subprocess.check_call(['a2dismod', 'remoteip'])
            if os.path.exists('/etc/apache2/conf-enabled/remoteip.conf'):
                subprocess.check_call(['a2disconf', 'remoteip'])
            LOG('Disabled remoteip logging')
        reload_required = True

    return reload_required


def configure_health_check(config=None):
    reload_required = False

    if not config:
        config = hookenv.config()

    if config.changed('enable_healthcheck'):
        LOG('Apache enable_healthcheck changed')
        if config['enable_healthcheck']:
            cmd = 'a2enmod'
            action = 'Enabled'
        else:
            cmd = 'a2dismod'
            action = 'Disabled'

        subprocess.check_call([cmd, 'cgid'])
        LOG('{} health check endpoint'.format(action))
        reload_required = True

    subprocess.call(['touch', HEALTH_CHECK_LOG])
    subprocess.call(['chown', HEALTH_CHECK_LOG_OWNER, HEALTH_CHECK_LOG])

    apache_path = unitdata.kv().get('apache-root')
    healthcheck_path = os.path.join(apache_path, 'health-check')
    host.mkdir(healthcheck_path, perms=0o655)
    healthcheck_script = os.path.join(healthcheck_path, 'health-check.py')
    with open('files/health_check.py', 'r') as f:
        content = f.read()
    host.write_file(path=healthcheck_script, content=content, perms=0o755)

    return reload_required


@util.run_as_user('root')
def render_configs():
    '''Render the configuration templates for apache'''

    reload_required = False

    LOG('Rendering apache2 configuration templates')
    config = hookenv.config()
    if config.changed('apache2_mpm_type'):
        reload_required = True
        if config['apache2_mpm_type'] == 'worker':
            LOG('Using apache2-mpm-worker')
            subprocess.check_call(['a2dismod', 'mpm_prefork'])
            subprocess.check_call(['a2enmod', 'mpm_worker'])
        elif config['apache2_mpm_type'] == 'prefork':
            LOG('Using apache2-mpm-prefork')
            # Package only in Trusty and below.
            subprocess.check_call(['a2dismod', 'mpm_worker'])
            subprocess.check_call(['a2enmod', 'mpm_prefork'])

    reload_required = create_metadata_site() or reload_required
    reload_required = create_mpm_workerfile() or reload_required
    reload_required = create_security() or reload_required
    reload_required = configure_remoteip_logging() or reload_required
    reload_required = configure_health_check() or reload_required

    start(start_only=(not reload_required))


def update_checks(nrpe_config):
    '''Update nagios check for apache serving archive metadata'''

    path_base = hookenv.config()['path-base']
    try:
        with open('/run/cloud-init/status.json') as status:
            status_json = json.load(status)
    except OSError:
        pass
    else:
        # try to find proof we're in Azure
        # in any version of the status json
        if any('DataSourceAzure' in x.get('datasource', '') for x in status_json.values()):
            # In azure, we rely on cloud services
            # instead of doing our own haproxy work
            nrpe_config.add_check(
                shortname='load-balancer',
                description='e2e check that regional load balancer is functioning',
                check_cmd=(
                    'check_http '
                    '--hostname ubuntu-archive-asm.trafficmanager.net '
                    '--url "/{}/" '
                    '--string "dists/" '
                    '--expect \"200\\ OK\" '
                    '--header \"Accept-Encoding:\\ \\*\"'
                ).format(path_base),
            )

    nrpe_config.add_check(
        shortname='apache',
        description='Apache2 serving archive metadata',
        check_cmd=(
            'check_http '
            '--hostname localhost '
            '--url "/{}/" '
            '--string "dists/" '
            '--expect \"200\\ OK\" '
            '--header \"Accept-Encoding:\\ \\*\"'
        ).format(path_base),
    )
