#!/usr/bin/python3
'''ubuntu-repository-cache charm

This 'hooks' source constitutes all of the code for handle charm hooks.
'''

import os
import pwd
import shutil

from charms import reactive
from charms.layer import status
from charmhelpers.core import hookenv, host, unitdata
from charmhelpers.contrib import unison

from lib.ubuntu_repository_cache import (
    metadata_sync,
    mirror,
    service,
    util,
)

LOG = hookenv.log
SERVICE = 'ubuntu-repository-cache'


@reactive.when_not('ubuntu-repository-cache.installed')
def install():
    '''Perform service installation'''

    status.maintenance('Installing %s' % SERVICE)
    reactive.clear_flag('ubuntu-repository-cache.active')
    service.install()
    reactive.set_flag('ubuntu-repository-cache.installed')
    # Needed to make sure any initial MPM tuning and mode will be made live.
    reactive.set_flag('ubuntu-repository-cache.apache2-restart-required')


@reactive.when('config.changed')
def config_changed():
    '''Handle charm configuration changes'''

    status.maintenance('Config changed for %s' % SERVICE)
    reactive.clear_flag('ubuntu-repository-cache.active')
    reactive.clear_flag('ubuntu-repository-cache.configured')
    reactive.clear_flag('nagios-nrpe.configured')


@reactive.when_any(
    'config.changed.apache2_mpm_type',
    'config.changed.apache2_mpm_threadlimit',
    'config.changed.apache2_mpm_serverlimit',
)
def apache2_restart_required():
    reactive.set_flag('ubuntu-repository-cache.apache2-restart-required')


@reactive.when('ubuntu-repository-cache.installed')
@reactive.when_not('ubuntu-repository-cache.configured')
def configure_services():
    service.render_configs()
    update_website_relation_data()

    # We should always do this anyway, but currently,
    # due to LP#1626318, following a reboot, the ephemeral
    # disk will be rendered immutable and read-only.
    azure_disk_warning = 'ephemeral-disk-warning'
    if host.service_available(azure_disk_warning):
        LOG('Disabling {}'.format(azure_disk_warning))
        host.service_pause(azure_disk_warning)

    # We want to ensure not all units rotate out and graceful
    # restart/reload apache at the same time.
    shutil.copy2('files/cron_random_sleep.sh', '/etc/cron.daily/000-cron-random-sleep')

    reactive.set_flag('ubuntu-repository-cache.configured')


@reactive.when('ubuntu-repository-cache.configured')
@reactive.when('ubuntu-repository-cache.started')
@reactive.when_not('ubuntu-repository-cache.active', 'ubuntu-repository-cache.apache2-restart-required')
def set_active(version_file='version'):
    revision = ''
    if os.path.exists(version_file):
        with open(version_file) as f:
            line = f.readline().strip().split('-')
        # We only want the first 8 characters, that's enough to tell
        # which version of the charm we're using.
        ver = line[-1][:8]
        # For bzr, we want to include the date as it's harder to determine
        # the date from the generated commit ID compared with git.
        if len(line) > 2:
            # Again, we only want the first 8 chars.
            date = line[-2][:8]
            revision = ' (source version/commit {}-{})'.format(date, ver)
        else:
            revision = ' (source version/commit {})'.format(ver)
    status.active('Ready{}'.format(revision))
    reactive.set_flag('ubuntu-repository-cache.active')


def update_website_relation_data():
    relation_id = None
    for relid in hookenv.relation_ids('website'):
        relation_id = relid
    if not relation_id:
        return

    # LP:1876029: Work around Juju not updating 'private-address'
    primary_address = hookenv.network_get_primary_address('website')
    LOG('Updating relation with primary IP address: {}'.format(primary_address))
    settings = {'private-address': primary_address, 'hostname': primary_address, 'port': 80}
    hookenv.relation_set(relation_id=relation_id, relation_settings=settings)


@reactive.when('ubuntu-repository-cache.configured')
@reactive.when('ubuntu-repository-cache.apache2-restart-required')
def apache2_restart():
    status.maintenance('Restarting apache2 as MPM settings have changed')
    reactive.clear_flag('ubuntu-repository-cache.active')
    host.service_restart('apache2')
    reactive.clear_flag('ubuntu-repository-cache.apache2-restart-required')


@reactive.hook('upgrade-charm')
def upgrade_charm():
    '''Handle charm updates'''

    status.maintenance('Upgrading %s' % SERVICE)
    reactive.clear_flag('ubuntu-repository-cache.active')
    reactive.clear_flag('ubuntu-repository-cache.configured')
    reactive.clear_flag('nagios-nrpe.configured')


@reactive.hook('start')
def start():
    '''Start the charm's services'''

    status.maintenance('Starting %s' % SERVICE)
    reactive.clear_flag('ubuntu-repository-cache.active')
    service.render_configs()

    sync_host = hookenv.config()['sync-host']
    mirror_series = hookenv.config()['mirror-series'].strip()
    config = hookenv.config()
    if config['sync-on-start']:
        status.maintenance('Performing initial sync from %s' % sync_host)
        # This creates a semaphore to prevent cronjobs form eventually running
        # while this is still syncing.
        try:
            with util.FileSemaphore(util.SEMAPHORE_FILE):
                meta_ver = metadata_sync.mirror_archive(
                    sync_host, unitdata.kv().get('apache-root'), hookenv.local_unit(), LOG, mirror_series
                )
        except util.SemaphoreExistsError:
            LOG(
                'Another metadata sync seems to be running (semaphore found '
                'at {}). Bailing out.'.format(util.SEMAPHORE_FILE)
            )
            raise

        if meta_ver:
            status.maintenance('Initial sync from {} completed ver={}.'.format(sync_host, meta_ver))
            service.start()
        else:
            status.maintenance('Leader could not gather new metadata', hookenv.WARNING)
    else:
        LOG('Skipping initial content sync, "sync-on-start" config ' 'option was not set')

    reactive.set_flag('ubuntu-repository-cache.started')


@reactive.hook('stop')
def stop():
    '''Stop the charm's services'''

    status.maintenance('Stopping %s' % SERVICE)
    service.stop()


@reactive.hook('website-relation-joined')
def website_relation_joined():
    '''Provide hostname/port for the website relationship'''

    LOG('Website relation joined for %s' % SERVICE)
    settings = {'hostname': hookenv.unit_private_ip(), 'port': 80}
    hookenv.relation_set(relation_settings=settings)


@reactive.hook('cluster-relation-joined')
def cluster_relation_joined():
    '''Join a peer cluster relationship'''

    LOG('Cluster relation joined for %s' % SERVICE)
    unison.ssh_authorized_peers(peer_interface='cluster', user='www-sync', group='www-data')
    # Gather our own set of SSH known hosts, we can't rely on the one
    # copied from the leader as it doesn't include itself - LP:1803453
    addresses = [
        hookenv.ingress_address(rid=u.rid, unit=u.unit) for u in hookenv.iter_units_for_relation_name('cluster')
    ]
    unison.write_known_hosts(user='www-sync', hosts=addresses)
    # LP:1770071: Work around charm-helper's unison ownership
    #             https://github.com/juju/charm-helpers/issues/487
    _fix_ssh_ownership(user='www-sync')
    mirror.peer_update_metadata()
    service.render_configs()


@reactive.hook('cluster-relation-changed')
def cluster_relation_changed():
    '''Handle peer cluster relationship changes'''

    LOG('Cluster relation changed for %s' % SERVICE)
    unison.ssh_authorized_peers(peer_interface='cluster', user='www-sync', group='www-data')
    # LP:1770071: Work around charm-helper's unison ownership
    #             https://github.com/juju/charm-helpers/issues/487
    _fix_ssh_ownership(user='www-sync')
    mirror.peer_update_metadata()
    service.render_configs()
    update_website_relation_data()


@reactive.hook('cluster-relation-departed')
def cluster_relation_departed():
    '''Handle peer cluster relationship departures'''

    LOG('Cluster relation departed for %s' % SERVICE)
    unison.ssh_authorized_peers(peer_interface='cluster', user='www-sync', group='www-data')
    # LP:1770071: Work around charm-helper's unison ownership
    #             https://github.com/juju/charm-helpers/issues/487
    _fix_ssh_ownership(user='www-sync')
    mirror.peer_update_metadata()
    service.render_configs()


@reactive.hook('cluster-relation-broken')
def cluster_relation_broken():
    '''Leave a peer cluster relationship'''

    LOG('Cluster relation broken for %s' % SERVICE)
    service.render_configs()


@reactive.hook('nrpe-external-master-relation-changed')
@reactive.hook('local-monitors-relation-changed')
def update_checks():
    reactive.clear_flag('nagios-nrpe.configured')


@reactive.when('ubuntu-repository-cache.configured')
@reactive.when('nrpe-external-master.available')
@reactive.when_not('nagios-nrpe.configured')
def configure_nagios():
    '''Update Nagios checks for all services'''

    service.update_checks()
    reactive.set_flag('nagios-nrpe.configured')


@reactive.hook('update-status')
def update_status():
    # Work around LP#1797297 and LP#1909569 and ensure the leader_id setting is correct.
    if not hookenv.is_leader():
        return
    local_unit = hookenv.local_unit()
    leader_id = hookenv.leader_get(attribute="leader_id")
    if leader_id != local_unit:
        LOG('Current leader is {} but leader_id is {}, updating to {}'.format(local_unit, leader_id, local_unit))
        hookenv.leader_set(leader_id=local_unit)


@reactive.hook('leader-elected')
def leader_elected():
    """
    The current unit just was elected leader of the cluster and will be "it"
    for at least 30 seconds.

    We need to tell other units what the unit name for the leader is (the unit
    name of the currently-running hook).
    """
    # Per LP#1723184, double check that we're the leader.
    if not hookenv.is_leader():
        LOG('leader-elected fired. This is not the leader')
        return
    LOG('leader-elected fired. This unit is the new leader: {}'.format(hookenv.local_unit()))
    hookenv.leader_set(leader_id=hookenv.local_unit())
    mirror.peer_update_metadata()
    service.render_configs()


@reactive.hook('leader-settings-changed')
def leader_settings_changed():
    """
    The leader settings changed. Most likely a new leader was elected and the
    leader_id was updated with leader_set.

    This simply logs the fact that it happened, for debugging purposes and to
    get a clearer picture of what happens.
    """

    leader_id = hookenv.leader_get(attribute="leader_id")
    LOG('leader-settings-changed fired. New leader_id: {}'.format(leader_id))
    mirror.peer_update_metadata()
    service.render_configs()
    update_website_relation_data()


# LP:1770071: Work around charm-helper's unison ownership
#             https://github.com/juju/charm-helpers/issues/487
def _fix_ssh_ownership(user='www-sync'):
    ssh_dir = os.path.join(os.path.expanduser('~{}'.format(user)), '.ssh')
    uid = pwd.getpwnam(user).pw_uid
    for f in ['authorized_keys', 'known_hosts']:
        fname = os.path.join(ssh_dir, f)
        if os.path.exists(fname):
            os.chown(fname, uid, -1)
