import os
import shutil
import sys
import tempfile
import unittest
from unittest import mock

# We also need to mock up charms.layer so we can run unit tests without having
# to build the charm and pull in layers such as layer-status.
sys.modules['charms.layer'] = mock.MagicMock()

from charms.layer import status  # NOQA: E402
from charmhelpers.core import hookenv, unitdata  # NOQA: E402

# Add path to where our reactive layer lives and import.
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))
from lib.ubuntu_repository_cache import apache  # NOQA: E402


class TestCharm(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.tmpdir = tempfile.mkdtemp(prefix='charm-unittests-')
        self.addCleanup(shutil.rmtree, self.tmpdir)
        os.environ['UNIT_STATE_DB'] = os.path.join(self.tmpdir, '.unit-state.db')

        unitdata.kv().set('apache-root', os.path.join(self.tmpdir, 'apache'))

        self.charm_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

        patcher = mock.patch('charmhelpers.core.hookenv.log')
        self.mock_log = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_log.return_value = ''

        patcher = mock.patch('charmhelpers.core.hookenv.charm_dir')
        self.mock_charm_dir = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_charm_dir.return_value = self.charm_dir

        patcher = mock.patch('charmhelpers.core.hookenv.local_unit')
        self.mock_local_unit = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_local_unit.return_value = 'mock-ubuntu-repository-cache/0'

        patcher = mock.patch('charmhelpers.core.host.log')
        self.mock_log = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_log.return_value = ''

        patcher = mock.patch('multiprocessing.cpu_count')
        self.mock_cpu_count = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_cpu_count.return_value = 2

        status.active.reset_mock()
        status.blocked.reset_mock()
        status.maintenance.reset_mock()
        # We also need to reset config.changed states.
        if os.path.exists('.juju-persistent-config'):
            os.unlink('.juju-persistent-config')

    @mock.patch('charmhelpers.core.host.write_file')
    @mock.patch('os.chown')
    @mock.patch('os.setegid')
    @mock.patch('os.seteuid')
    @mock.patch('subprocess.call')
    @mock.patch('subprocess.check_call')
    def test_configure_health_check(self, check_call, call, seteuid, setegid, chown, write_file):
        config = hookenv.Config({'enable_healthcheck': True})
        self.assertTrue(apache.configure_health_check(config))
        check_call.assert_called_with(['a2enmod', 'cgid'])

        check_call.reset_mock()
        config = hookenv.Config({'enable_healthcheck': False})
        self.assertTrue(apache.configure_health_check(config))
        check_call.assert_called_with(['a2dismod', 'cgid'])

        # No change
        check_call.reset_mock()
        config = hookenv.Config({'enable_healthcheck': True})
        config.save()
        config = hookenv.Config({'enable_healthcheck': True})
        self.assertFalse(apache.configure_health_check(config))

    @mock.patch('lib.ubuntu_repository_cache.apache.start')
    @mock.patch('lib.ubuntu_repository_cache.util.get_failover_host')
    @mock.patch('os.chown')
    @mock.patch('os.fchown')
    @mock.patch('subprocess.call')
    def test_create_metadata_site(self, call, fchown, chown, get_failover_host, start):
        get_failover_host.return_value = '10.1.1.1'
        config = hookenv.Config(
            {
                'display-host': 'archive.ubuntu.com',
                'enable_healthcheck': True,
                'path-base': 'ubuntu',
                'sync-host': 'us.archive.ubuntu.com',
            }
        )
        self.assertTrue(apache.create_metadata_site(config, apache2_conf_path=self.tmpdir))
        call.assert_called_with(['a2ensite', 'archive_ubuntu_com'])
        chown.assert_called()

        # No change, no need to reload
        call.reset_mock()
        config.save()
        config = hookenv.Config(
            {
                'display-host': 'archive.ubuntu.com',
                'enable_healthcheck': True,
                'path-base': 'ubuntu',
                'sync-host': 'us.archive.ubuntu.com',
            }
        )
        self.assertFalse(apache.create_metadata_site(config, apache2_conf_path=self.tmpdir))
        call.assert_not_called()

    @mock.patch('charmhelpers.core.templating.render')
    @mock.patch('lib.ubuntu_repository_cache.apache.start')
    @mock.patch('lib.ubuntu_repository_cache.util.get_failover_host')
    @mock.patch('lib.ubuntu_repository_cache.util.in_failover')
    @mock.patch('subprocess.call')
    def test_create_metadata_site_in_failover(self, call, in_failover, get_failover_host, start, render):
        config = hookenv.Config(
            {
                'display-host': 'archive.ubuntu.com',
                'enable_healthcheck': True,
                'path-base': 'ubuntu',
                'sync-host': 'us.archive.ubuntu.com',
            }
        )
        get_failover_host.return_value = '10.1.1.1'
        in_failover.return_value = True

        os.mkdir(os.path.join(self.tmpdir, 'apache'))

        apache.create_metadata_site(config)
        render.assert_called_with(
            'apache2/archive_ubuntu_com.conf',
            '/etc/apache2/sites-available/archive_ubuntu_com.conf',
            {
                'DocumentRoot': os.path.join(self.tmpdir, 'apache'),
                'Sync_Host': 'us.archive.ubuntu.com',
                'Display_Host': 'archive.ubuntu.com',
                'Path_Base': 'ubuntu',
                'MetadataProxy': 'ProxyRemote      http://archive.ubuntu.com/ http://10.1.1.1:80\n'
                + '\tProxyPass        / http://archive.ubuntu.com/\n'
                + '\tProxyPassReverse / http://archive.ubuntu.com/',
                'Enable_Healthcheck': True,
            },
        )
        call.assert_called_with(['a2ensite', 'archive_ubuntu_com'])

    @mock.patch('charmhelpers.core.templating.render')
    @mock.patch('lib.ubuntu_repository_cache.apache.start')
    @mock.patch('lib.ubuntu_repository_cache.util.get_failover_host')
    @mock.patch('subprocess.call')
    def test_create_metadata_site_no_failover(self, call, get_failover_host, start, render):
        config = hookenv.Config(
            {
                'display-host': 'archive.ubuntu.com',
                'enable_healthcheck': True,
                'path-base': 'ubuntu',
                'sync-host': 'us.archive.ubuntu.com',
            }
        )

        os.mkdir(os.path.join(self.tmpdir, 'apache'))

        apache.create_metadata_site(config)
        render.assert_called_with(
            'apache2/archive_ubuntu_com.conf',
            '/etc/apache2/sites-available/archive_ubuntu_com.conf',
            {
                'DocumentRoot': os.path.join(self.tmpdir, 'apache'),
                'Sync_Host': 'us.archive.ubuntu.com',
                'Display_Host': 'archive.ubuntu.com',
                'Path_Base': 'ubuntu',
                'MetadataProxy': '# No failover configured',
                'Enable_Healthcheck': True,
            },
        )
        call.assert_called_with(['a2ensite', 'archive_ubuntu_com'])

    @mock.patch('charmhelpers.core.host.service_start')
    @mock.patch('charmhelpers.core.host.service_reload')
    @mock.patch('charmhelpers.core.host.service_running')
    # For @util.run_as_user()
    @mock.patch('os.setegid')
    @mock.patch('os.seteuid')
    def test_start(self, seteuid, setegid, service_running, service_reload, service_start):
        service_running.return_value = True
        apache.start()
        service_reload.assert_called_with('apache2')
        service_start.assert_not_called()

        service_reload.reset_mock()
        service_start.reset_mock()
        service_running.return_value = False
        apache.start()
        service_reload.assert_not_called()
        service_start.assert_called_with('apache2')

        service_reload.reset_mock()
        service_start.reset_mock()
        service_running.return_value = True
        apache.start(start_only=True)
        service_reload.assert_not_called()
        service_start.assert_not_called()

        service_reload.reset_mock()
        service_start.reset_mock()
        service_running.return_value = False
        apache.start(start_only=True)
        service_reload.assert_not_called()
        service_start.assert_called_with('apache2')

    @mock.patch('charmhelpers.core.templating.render')
    @mock.patch('os.symlink')
    @mock.patch('subprocess.check_call')
    def test_create_mpm_workerfile(self, check_call, symlink, render):
        config = hookenv.Config(
            {
                'apache2_mpm_maxconnectionsperchild': 10000,
                'apache2_mpm_maxrequestworkers': 0,
                'apache2_mpm_maxsparethreads': 0,
                'apache2_mpm_minsparethreads': 0,
                'apache2_mpm_serverlimit': 0,
                'apache2_mpm_startservers': 0,
                'apache2_mpm_threadlimit': 0,
                'apache2_mpm_threadsperchild': 0,
            }
        )
        self.assertTrue(apache.create_mpm_workerfile(config))
        render.assert_called_with(
            'apache2/mpm_worker.template',
            '/etc/apache2/conf-available/000mpm-worker.conf',
            {
                'maxconnectionsperchild': 10000,
                'maxrequestworkers': 1024,
                'maxsparethreads': 1024,
                'minsparethreads': 512,
                'serverlimit': 2,
                'startservers': 1,
                'threadlimit': 512,
                'threadsperchild': 512,
            },
        )
        check_call.assert_called_with(['a2enconf', '000mpm-worker'])
        symlink.assert_called_with(
            '/etc/apache2/conf-available/000mpm-worker.conf', '/etc/apache2/mods-available/mpm_worker.conf'
        )

        # No change, no need to reload
        render.reset_mock()
        check_call.reset_mock()
        symlink.reset_mock()
        config.save()
        config = hookenv.Config(
            {
                'apache2_mpm_maxconnectionsperchild': 10000,
                'apache2_mpm_maxrequestworkers': 0,
                'apache2_mpm_maxsparethreads': 0,
                'apache2_mpm_minsparethreads': 0,
                'apache2_mpm_serverlimit': 0,
                'apache2_mpm_startservers': 0,
                'apache2_mpm_threadlimit': 0,
                'apache2_mpm_threadsperchild': 0,
            }
        )
        self.assertFalse(apache.create_mpm_workerfile(config))

    @mock.patch('charmhelpers.core.templating.render')
    @mock.patch('subprocess.check_call')
    def test_create_security(self, check_call, render):
        config = hookenv.Config(
            {
                'apache2_server_tokens': 'OS',
                'apache2_server_signature': 'On',
                'apache2_trace_enabled': 'Off',
            }
        )
        self.assertTrue(apache.create_security(config))
        render.assert_called_with(
            'apache2/security.template',
            '/etc/apache2/conf-available/security.conf',
            {
                'server_tokens': 'OS',
                'server_signature': 'On',
                'trace_enabled': 'Off',
            },
        )
        check_call.assert_called_with(['a2enconf', 'security'])

        # No change, no need to reload
        render.reset_mock()
        check_call.reset_mock()
        config.save()
        config = hookenv.Config(
            {
                'apache2_server_tokens': 'OS',
                'apache2_server_signature': 'On',
                'apache2_trace_enabled': 'Off',
            }
        )
        self.assertFalse(apache.create_security(config))

    @mock.patch('os.chown')
    @mock.patch('os.fchown')
    # For @util.run_as_user()
    @mock.patch('os.setegid')
    @mock.patch('os.seteuid')
    @mock.patch('subprocess.check_call')
    def test_configure_remoteip_logging(self, check_call, seteuid, setegid, fchown, chown):
        config = hookenv.Config({'remoteip_logging': True})
        self.assertTrue(apache.configure_remoteip_logging(config, apache2_conf_path=self.tmpdir))
        check_call.assert_called_with(['a2enmod', 'remoteip'])
        with open(os.path.join(self.tmpdir, 'conf-available', 'remoteip.conf'), 'r') as f:
            got = f.read()
        with open('templates/apache2/remoteip.conf', 'r') as f:
            want = f.read().strip()
        self.assertEqual(want, got)

        # No change
        config.save()
        config = hookenv.Config({'remoteip_logging': True})
        self.assertFalse(apache.configure_remoteip_logging(config, apache2_conf_path=self.tmpdir))

        check_call.reset_mock()
        config = hookenv.Config({'remoteip_logging': False})
        self.assertTrue(apache.configure_remoteip_logging(config, apache2_conf_path=self.tmpdir))
        check_call.assert_called_with(['a2dismod', 'remoteip'])
