import os
import shutil
import sys
import tempfile
import unittest
from unittest import mock

# Add path to where our reactive layer lives and import.
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))
from files import health_check  # NOQA: E402


class TestCharm(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.tmpdir = tempfile.mkdtemp(prefix='charm-unittests-')
        self.addCleanup(shutil.rmtree, self.tmpdir)

        self.charm_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

    def test_generate_uuid(self):
        uuid4 = b'\x8e\x91w\x01\x17\xdeEa\xa9\xb3\xf4Bk\x16\xc5L'

        want = 'jpF3ARfe'
        got = health_check.generate_uuid(uuid4)
        self.assertEqual(want, got)

    def test_render_influx(self):
        metric_name = 'mymetric'
        labels = 'label1=mylabel1,label2=mylabel2'
        value = '0.1'

        want = 'mymetric,label1=mylabel1,label2=mylabel2 value=0.1'
        got = health_check.render_influx(metric_name, labels, value)
        self.assertEqual(want, got)

        want = 'mymetric value=0.1'
        got = health_check.render_influx(metric_name, '', value)
        self.assertEqual(want, got)

        want = 'mymetric value=0.1'
        got = health_check.render_influx(metric_name, None, value)
        self.assertEqual(want, got)

    @mock.patch('logging.warning')
    def test_send_to_influx(self, log_warning):
        metrics = 'mymetric,label1=mylabel1,label2=mylabel2 value=0.1'

        send_to = '127.0.0.1:65530'
        health_check.send_to_influx(metrics, send_to)
        log_warning.assert_called()

        log_warning.reset_mock()
        with mock.patch('socket.socket.connect'):
            with mock.patch('socket.socket.makefile'):
                health_check.send_to_influx(metrics, send_to)
                log_warning.assert_not_called()

    @mock.patch('files.health_check.send_to_influx')
    @mock.patch('logging.info')
    def test_health_check_disabled(self, log_info, send_to_influx):
        disabled_flag = os.path.join(self.tmpdir, 'health-check-disabled.flag')
        os.mknod(disabled_flag)
        os.utime(disabled_flag, times=(1613518897.123, 1613518897.123))
        health_check.main(disabled_flag)
        log_info.assert_called_with('DISABLED since %.3f', 1613518897.123)
        send_to_influx.assert_called_with('ubuntu_repository_cache_health,status=DISABLED value=5.03')

    @mock.patch('files.health_check.send_to_influx')
    @mock.patch('logging.warning')
    def test_health_check_missing_metadata(self, log_warning, send_to_influx):
        health_check.main()
        log_warning.assert_called_with('DISABLED ubuntu_active missing')
        send_to_influx.assert_called_with('ubuntu_repository_cache_health,status=DISABLED value=4.04')
